/*
 *
 *
 *                      JACQUET Florent -- Skia
 *                      florent.jacquet@utbm.fr
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <semaphore.h>
#include <pthread.h>
#include <time.h>


#include <crossroad.h>
#include <map.h>
#include <v_controller.h>
#include <utils.h>

Crossroad *new_crossroad() {
	static int id_counter = 0;
	Crossroad *c = malloc(sizeof(Crossroad));
	c->id = id_counter++;
	c->output_nb = 0;
	c->input_nb = 0;
	c->load = 0;
	if ((c->output = malloc(MAX_OUTPUT*sizeof(Road*))) == NULL) goto fail;
	if ((c->input = malloc(MAX_INPUT*sizeof(Road*))) == NULL) goto fail1;
	return c;
fail1:
	free(c->output);
fail:
	free(c);
	if (DEBUG) fprintf(stderr, "DEBUG: Can not create crossroad\n");
	return NULL;
}

void free_crossroad(Crossroad *c) {
	free(c->output);
	free(c->input);
	free(c);
}

void add_to_crossroad(Crossroad *c, Road *r, int pos, int input) {
	if (input == INPUT) {
		c->input[pos] = r;
		r->end = c;
	} else {
		c->output[pos] = r;
		r->start = c;
	}
}

static int get_direction(Crossroad *c, Vehicle *v) {
	int c_id = c->id;
	int goal = v->goal;

	if (c->load < 30 || goal/10 == c_id) {
		return (goal%10)%4;
	} else {
		return 3-(c_id+2)%4;
	}
}

void *start_crossroad(void *a) {
	void **arg = (void**)a;
	int t_nb = (int)arg[1];
	Map *m = (Map*)arg[2];
	Crossroad *c = m->crossroads[t_nb];
	Vehicle *buf[4];
	int stored_v = 0;
	int i = 0;

	while (run) {
		stored_v = 0;
		for (i=0; i < 4; i++) {
			sem_wait(m->road_mut[c->input[i]->id]);
			Vehicle *tmp = pop_from_road(c->input[i]);
			if (tmp != NULL)
				buf[stored_v++] = tmp;
			sem_post(m->road_mut[c->input[i]->id]);
		}

		for (i=0; i < stored_v; i++) {
			int idx = get_direction(c, buf[i]);
			Road *r = c->output[idx];
			sem_wait(m->road_mut[r->id]);
			while (add_to_road(r, buf[i]) == -1) {
				sem_post(m->road_mut[r->id]);
				usleep(CLOCK);
				sem_wait(m->road_mut[r->id]);
			}
			sem_post(m->road_mut[r->id]);
		}

		c->load = 0;
		for (i=0; i < 4; i++) {
			sem_wait(m->road_mut[c->input[i]->id]);
			c->load += c->input[i]->load;
			sem_post(m->road_mut[c->input[i]->id]);
		}

		/* Pop X vehicle to store them
		 * 	-> Need of a Vehicle *pop(Road*) function!
		 * Send all the stored vehicle to the correct road
		 * Check if the crossroad is overcharged
		 *		if so, slow down the vehicle 
		 */
		usleep(6*CLOCK);
	}
	return NULL;
}


