/*
 *
 *
 *                      JACQUET Florent -- Skia
 *                      florent.jacquet@utbm.fr
 */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <semaphore.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>



#include <utils.h>
#include <map.h>
#include <v_controller.h>
#include <crossroad.h>
#include <print.h>


void stopping(int signo, siginfo_t *sinfo, void *useless);
void **context;

int main(int argc, char** argv)
{
	void **arg;
	sem_t *mut;
	int num = 0;
	pthread_t pth_id[7];
	struct sigaction stop;
	Map *map = NULL;

	/* Map */
	map = new_map();

	/* Map mutex */
	mut = sem_open("map", O_RDWR | O_CREAT, 0666, 1);
	if (mut == SEM_FAILED) {
		fprintf(stderr, "DEBUG: Semaphore error\n");
		exit(EXIT_FAILURE);
	}


	arg = malloc(sizeof(void*)*3);
	arg[0] = (void*)mut;
	arg[2] = (void*)map;

	/* Launch threads */
	run = 1;

	/* The crossroads */
	arg[1] = (void*)num;
	if (pthread_create(pth_id + num, 0, start_crossroad, (void*)arg) == 0) {
		printf("Main: thread number %d created: id = %d\n", num,(int)pth_id[num]);
		num++;
	} else fprintf(stderr, "Thread error\n");
	usleep(CLOCK);
	arg[1] = (void*)num;
	if (pthread_create(pth_id + num, 0, start_crossroad, (void*)arg) == 0) {
		printf("Main: thread number %d created: id = %d\n", num,(int)pth_id[num]);
		num++;
	} else fprintf(stderr, "Thread error\n");
	usleep(CLOCK);
	arg[1] = (void*)num;
	if (pthread_create(pth_id + num, 0, start_crossroad, (void*)arg) == 0) {
		printf("Main: thread number %d created: id = %d\n", num,(int)pth_id[num]);
		num++;
	} else fprintf(stderr, "Thread error\n");
	usleep(CLOCK);
	arg[1] = (void*)num;
	if (pthread_create(pth_id + num, 0, start_crossroad, (void*)arg) == 0) {
		printf("Main: thread number %d created: id = %d\n", num,(int)pth_id[num]);
		num++;
	} else fprintf(stderr, "Thread error\n");

	map->fleet_nb = 0;
	/* Then the vehicle controllers */
	arg[1] = (void*)CAR;
	if (pthread_create(pth_id + num, 0, start_controller, (void*)arg) == 0) {
		printf("Main: thread number %d created: id = %d\n", num,(int)pth_id[num]);
		num++;
	} else fprintf(stderr, "Thread error\n");
	usleep(CLOCK);
	arg[1] = (void*)TRUCK;
	if (pthread_create(pth_id + num, 0, start_controller, (void*)arg) == 0) {
		printf("Main: thread number %d created: id = %d\n", num,(int)pth_id[num]);
		num++;
	} else fprintf(stderr, "Thread error\n");
	usleep(CLOCK);
	arg[1] = (void*)MOTOBIKE;
	if (pthread_create(pth_id + num, 0, start_controller, (void*)arg) == 0) {
		printf("Main: thread number %d created: id = %d\n", num,(int)pth_id[num]);
		num++;
	} else fprintf(stderr, "Thread error\n");

	/* Context definition */
	context = malloc(sizeof(void*)*5);
	context[0] = (void*)pth_id;
	context[1] = (void*)map;
	context[2] = (void*)num;
	context[3] = (void*)mut;
	context[4] = (void*)arg;

	/* Signal handling */
	memset(&stop, 0, sizeof(struct sigaction));
	sigemptyset(&stop.sa_mask);
	stop.sa_sigaction = stopping;
	stop.sa_flags = SA_SIGINFO;
	sigaction(SIGINT, &stop, NULL);

	while(run) {
		print_map(map);
		usleep(CLOCK);
	}

/*
    for (i = 0; i < num; i++) {
    	if (pthread_join(pth_id[i], (void**)&ret) == 0)
		printf ("Joined with thread %ld, returned value was %ld\n", (long) pth_id[i], (long) ret);
    }
*/

    return 0;
}

char *get_error(int err) {
	char *str = malloc(8*sizeof(char));
	switch (err) {
		case ESRCH:
			strcpy(str, "ESRCH");
			break;
		case EINVAL:
			strcpy(str, "EINVAL");
			break;
		case EDEADLK:
			strcpy(str, "EDEADLK");
			break;
		default:
			strcpy(str, "TROLL");
	}
	return str;
}

void stopping(int signo, siginfo_t *sinfo, void *useless) {
	pthread_t *pth_id = (pthread_t*) context[0];
	Map *map = (Map*) context[1];
	int th_nb = (int) context[2];
	int i = 0;
	void **ret;

	ret = malloc(th_nb*sizeof(void*));
	sem_t *mut = context[3];
	if (DEBUG) fprintf (stderr, "DEBUG: Map: %p\n", map);
	run = 0;
	for (; i<th_nb; i++) {
		int err = 0;
		char *tmp = NULL;
		if((err=pthread_join(pth_id[i], ret+i)) != 0 && DEBUG) {
			fprintf(stderr, "DEBUG: Thread %d was not joined (error: %s)\n", (int)pth_id[i], tmp=get_error(err));
			free(tmp);
		} else printf("Thread %d terminated with status %d\n", i, (int)ret[i]);
	}
	sem_wait(mut);
	free(ret);
	free_map(map);
	sem_post(mut);
	sem_close(mut);
	free(context[4]); /* arg */
	free(context);
	exit(EXIT_SUCCESS);
}
