/*
 *
 *
 *                      JACQUET Florent -- Skia
 *                      florent.jacquet@utbm.fr
 */

#include <stdlib.h>
#include <stdio.h>
#include <semaphore.h>
#include <fcntl.h>

#include <utils.h>
#include <map.h>
#include <v_controller.h>
#include <crossroad.h>



Road *new_road(int width, int len, int input) {
	static int id_counter = 0;
	int i = 0, j = 0;
	Road *r = NULL;
	if((r = malloc(sizeof(Road))) == NULL) goto fail;
	r->id = id_counter++;
	r->length = len;
	r->width = (width>MAX_LANE) ? MAX_LANE : width;
	r->start = NULL;
	r->end = NULL;
	r->input = input;
	r->load = 0;
	if((r->lane = malloc(width*sizeof(void*))) == NULL) goto fail1;
	for(;i<width;i++) {
		if((r->lane[i] = malloc(len*sizeof(Vehicle*))) == NULL) goto fail2;
	}
	for(i=0;i<width;i++) {
		for(j=0;j<len;j++) {
			r->lane[i][j] = NULL;
		}
	}
	return r;
fail2:
	i--;
	for(;i>=0;i--)
		free(r->lane[i]);
fail1:
	free(r);
fail:
	if(DEBUG)
		fprintf(stderr, "DEBUG: Road creation failed\n");
	return NULL;
}

void free_road(Road *r) {
	int i = 0;
	for (; i < r->width; i++) {
		free(r->lane[i]);
	}
	free(r->lane);
	free(r);
	if (DEBUG) fprintf(stderr, "DEBUG: Road freed\n");
}

int add_to_road(Road *r, Vehicle *v) {
	int i = 0;
	while (i < r->width) {
		if (r->lane[i][0] == NULL) {
			v->offset = 0;
			r->lane[i][0] = v;
			v->road = r;
			v->lane = i;
			r->load++;
			return i;
		}
		i++;
	}
	if (DEBUG)
		fprintf(stderr, "DEBUG: Road full\n");
	return -1;
}

Vehicle *pop_from_road(Road *r) {
	Vehicle *v = NULL;
	int i = 0;
	for (i=0; i < r->width; i++) {
		if (r->lane[i][r->length-1] != NULL) {
			if (v == NULL) {
				v = r->lane[i][r->length-1];
				r->lane[i][r->length-1] = NULL;
			}/* else if (v->speed > r->lane[i][r->length-1]->speed) {
				v = r->lane[i][r->length-1];
				r->lane[i][r->length-1] = NULL;
			} */
		}
	}
	r->load -= (v!=NULL) ? 1 : 0;
	return v;
}

/* This is where the map is built */
Map *new_map() {
	Map *map = NULL;
	int i = 0;
	if ((map = malloc(sizeof(Map))) == NULL) goto fail;

	map->roads = malloc(24*sizeof(Road*));
	map->roads[i++] = new_road(3, ROAD_LENGTH, INPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, INPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, INPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, INPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, INPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, INPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, INPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, INPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(4, ROAD_LENGTH, OUTPUT);
	map->roads[i++] = new_road(3, ROAD_LENGTH, OUTPUT);
	map->road_nb = i;

	if ((map->road_mut = malloc(i*sizeof(sem_t*))) == NULL) goto fail1;
	for (i = 0; i < map->road_nb; i++) {
		char str[8] = {'r', 'o', 'a', 'd', i/100+'0', i/10+'0', i%10+'0', '\0'};
		map->road_mut[i] = sem_open(str, O_RDWR | O_CREAT, 0666, 1);
		if (map->road_mut[i] == SEM_FAILED) {
			fprintf(stderr, "DEBUG: Mutex creation error\n");
			exit(EXIT_FAILURE); /* Should not stop without freeing */
		}
	}

	if ((map->crossroads = malloc(4*sizeof(Crossroad*))) == NULL) goto fail2;
	i = 0;
	map->crossroads[i++] = new_crossroad();
	map->crossroads[i++] = new_crossroad();
	map->crossroads[i++] = new_crossroad();
	map->crossroads[i++] = new_crossroad();
	map->crossroad_nb = i;

	i = 0;
	add_to_crossroad(map->crossroads[0], map->roads[i++], UP,    INPUT);
	add_to_crossroad(map->crossroads[0], map->roads[i++], LEFT,  INPUT);
	add_to_crossroad(map->crossroads[1], map->roads[i++], UP,    INPUT);
	add_to_crossroad(map->crossroads[1], map->roads[i++], RIGHT, INPUT);
	add_to_crossroad(map->crossroads[2], map->roads[i++], LEFT,  INPUT);
	add_to_crossroad(map->crossroads[2], map->roads[i++], DOWN,  INPUT);
	add_to_crossroad(map->crossroads[3], map->roads[i++], RIGHT, INPUT);
	add_to_crossroad(map->crossroads[3], map->roads[i++], DOWN,  INPUT);

	add_to_crossroad(map->crossroads[0], map->roads[i++], UP,    OUTPUT);
	add_to_crossroad(map->crossroads[0], map->roads[i++], LEFT,  OUTPUT);
	add_to_crossroad(map->crossroads[1], map->roads[i++], UP,    OUTPUT);
	add_to_crossroad(map->crossroads[1], map->roads[i++], RIGHT, OUTPUT);
	add_to_crossroad(map->crossroads[2], map->roads[i++], LEFT,  OUTPUT);
	add_to_crossroad(map->crossroads[2], map->roads[i++], DOWN,  OUTPUT);
	add_to_crossroad(map->crossroads[3], map->roads[i++], RIGHT, OUTPUT);
	add_to_crossroad(map->crossroads[3], map->roads[i++], DOWN,  OUTPUT);

	add_to_crossroad(map->crossroads[0], map->roads[i],   DOWN,  OUTPUT);
	add_to_crossroad(map->crossroads[2], map->roads[i++], UP,     INPUT);
	add_to_crossroad(map->crossroads[0], map->roads[i],   DOWN,   INPUT);
	add_to_crossroad(map->crossroads[2], map->roads[i++], UP,    OUTPUT);
	add_to_crossroad(map->crossroads[0], map->roads[i],   RIGHT, OUTPUT);
	add_to_crossroad(map->crossroads[1], map->roads[i++], LEFT,   INPUT);
	add_to_crossroad(map->crossroads[0], map->roads[i],   RIGHT,  INPUT);
	add_to_crossroad(map->crossroads[1], map->roads[i++], LEFT,  OUTPUT);
	add_to_crossroad(map->crossroads[1], map->roads[i],   DOWN,  OUTPUT);
	add_to_crossroad(map->crossroads[3], map->roads[i++], UP,     INPUT);
	add_to_crossroad(map->crossroads[1], map->roads[i],   DOWN,   INPUT);
	add_to_crossroad(map->crossroads[3], map->roads[i++], UP,    OUTPUT);
	add_to_crossroad(map->crossroads[2], map->roads[i],   RIGHT, OUTPUT);
	add_to_crossroad(map->crossroads[3], map->roads[i++], LEFT,   INPUT);
	add_to_crossroad(map->crossroads[2], map->roads[i],   RIGHT,  INPUT);
	add_to_crossroad(map->crossroads[3], map->roads[i++], LEFT,  OUTPUT);


	if ((map->fleets = malloc(3*sizeof(Fleet*))) == NULL) goto fail3;
	map->fleet_nb = 0;

	return map;
fail3:
	free(map->crossroads);
fail2:
	free(map->road_mut);
fail1:
	free(map);
fail:
	return NULL;
}

void free_map(Map *m) {
	int i;
	for (i = 0; i < m->fleet_nb; i++)
		free_fleet(m->fleets[i]);
	free(m->fleets);
	if (DEBUG) fprintf(stderr, "DEBUG: Fleets freed\n");
	for (i = 0; i < m->road_nb; i++) {
		free_road(m->roads[i]);
		sem_close(m->road_mut[i]);
	}
	free(m->road_mut);
	free(m->roads);
	if (DEBUG) fprintf(stderr, "DEBUG: Roads freed\n");
	for (i = 0; i < m->crossroad_nb; i++)
		free_crossroad(m->crossroads[i]);
	free(m->crossroads);
	if (DEBUG) fprintf(stderr, "DEBUG: Crossroads freed\n");
	free(m);
	if (DEBUG) fprintf(stderr, "DEBUG: Map freed\n");
}



