/*
 *
 *
 *                      JACQUET Florent -- Skia
 *                      florent.jacquet@utbm.fr
 */

#include <stdio.h>
#include <stdlib.h>
#include <semaphore.h>
#include <fcntl.h>
#include <string.h>

#include <print.h>


#define RESET   "\033[0m"
#define CYAN    "\033[36m"


void print_map(Map *m) {
	int i = 0, j=0;
	char **output;
	int chunk_size = 59;
	static int iteration = 0;
	int motos, cars, trucks;

	output = malloc(chunk_size*sizeof(char*));
	for (i=0;i<chunk_size;i++) {
		output[i] = malloc(chunk_size*sizeof(char));
	}
	for (i=0;i<chunk_size;i++) {
		for (j=0;j<chunk_size;j++) {
			output[i][j] = ' ';
		}
	}

	sprintf(output[chunk_size-3]+1, "Stats         ");
	sprintf(output[chunk_size-2]+1, "(approx)    ");

	/* Take the roads */
	for (i=0; i < m->road_nb; i++) {
		sem_wait(m->road_mut[m->roads[i]->id]);
	}

	/* Print everything */
	print_crossroad(output, 0, 0, m->crossroads[0]);
	print_crossroad(output, ROAD_LENGTH+MAX_LANE*2+3, 0, m->crossroads[1]);
	print_crossroad(output, 0, ROAD_LENGTH+MAX_LANE*2+3, m->crossroads[2]);
	print_crossroad(output, ROAD_LENGTH+MAX_LANE*2+3, ROAD_LENGTH+MAX_LANE*2+3, m->crossroads[3]);

	/* Compute the vehicle number */
	motos = cars = trucks = 0;
	for (i=0;i<chunk_size-1;i++) {
		for (j=0;j<chunk_size;j++) {
			char c = output[i][j];
			if (c == '@') printf (CYAN);
			printf("%c", c);
			if (c == '@') printf (RESET);
			if (c == '%') motos++;
			else if (c == 'x') cars++;
			else if (c == '#') trucks++;
		}
		printf("\n");
	}

	/* Release the roads */
	for (i=0; i < m->road_nb; i++) {
		sem_post(m->road_mut[m->roads[i]->id]);
	}

	sprintf(output[chunk_size-1]+1, "It %d, vehicle number: %d, motos:%d/cars:%d/trucks:%d", iteration++, motos+cars+trucks, motos, cars, trucks);
	printf("%s\n", output[chunk_size-1]);
	for (i=0;i<chunk_size;i++) {
		free(output[i]);
	}
	free(output);
}


void print_crossroad(char **output, int x, int y, Crossroad *c) {
	int r_len = 0;
	int width = 2*MAX_LANE+2;
	int i;
	if (c != NULL && c->input != NULL) r_len = c->input[0]->length;
	sprintf(output[y+r_len+width/2-1]+x+r_len+width/2-3, " load   ");
	sprintf(output[y+r_len+width/2]+x+r_len+width/2-2, " %03d   ", c->load);
	output[y+r_len+0][x+r_len+0] = output[y+r_len+0][x+r_len+width] =
		output[y+r_len+width][x+r_len+width] = output[y+r_len+width][x+r_len+0] = '*';
	for (i=r_len+1;i<r_len+width;i++) {
		output[y+r_len][x+i] = '-';
		output[y+r_len+width][x+i] = '-';
		output[y+i][x+r_len] = '|';
		output[y+i][x+r_len+width] = '|';
	}
	print_road(output, x+r_len+width/2+1, y, c->output[UP], VERTICAL, REVERSE);
	print_road(output, x+r_len+1, y+r_len+width+1, c->output[DOWN], VERTICAL, !REVERSE);
	print_road(output, x, y+r_len+1, c->output[LEFT], !VERTICAL, REVERSE);
	print_road(output, x+r_len+width+1, y+r_len+width/2+1, c->output[RIGHT], !VERTICAL, !REVERSE);

	print_road(output, x+r_len+1, y, c->input[UP], VERTICAL, !REVERSE);
	print_road(output, x+r_len+width/2+1, y+r_len+width+1, c->input[DOWN], VERTICAL, REVERSE);
	print_road(output, x, y+r_len+width/2+1, c->input[LEFT], !VERTICAL, !REVERSE);
	print_road(output, x+r_len+width+1, y+r_len+1, c->input[RIGHT], !VERTICAL, REVERSE);

}


static void print_case(char *output, Vehicle *v, int vertical) {
	if(v != NULL) {
		switch(v->type) {
			case TRUCK:
				*output = '#';
				break;
			case MOTOBIKE:
				*output = '%';
				break;
			default:
				*output = 'x';
		}
		if (v->id%30 == 0)
			*output = '@';
	} else {
		*output = (vertical) ? '\'' : '-';
	}
}

void print_road(char **output, int x, int y, Road *r, int vertical, int reverse) {
	int i = 0, j = 0;
	int w = r->width, l = r->length;

	if (vertical) { /* Vertical */
		if (reverse) { /* To the top */
			for(i=0;i<w;i++) {
				for(j=0;j<l;j++) {
					print_case(output[y+l-j-1]+x+w-i-1, r->lane[i][j], vertical);
				}
			}
		} else { /* To the bottom */
			for(i=0;i<w;i++) {
				for(j=0;j<l;j++) {
					print_case(output[y+j]+x+i, r->lane[i][j], vertical);
				}
			}
		}
	} else { /* Horizontal */
		if (reverse) { /* To the left */
			for(i=0;i<w;i++) {
				for(j=0;j<l;j++) {
					print_case(output[y+w-i-1]+x+l-j-1, r->lane[i][j], vertical);
				}
			}
		} else { /* To the right */
			for(i=0;i<w;i++) {
				for(j=0;j<l;j++) {
					print_case(output[y+i]+x+j, r->lane[i][j], vertical);
				}
			}
		}
	}
}



